#  Quapp Platform Project
#  qiskit_handler_factory.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.config.logging_config import job_logger
from quapp_common.factory.handler_factory import HandlerFactory
from quapp_common.handler.handler import Handler

from ..handler.invocation_handler import InvocationHandler
from ..handler.job_fetching_handler import JobFetchingHandler


class QiskitHandlerFactory(HandlerFactory):

    @staticmethod
    def create_handler(event, circuit_preparation_fn,
            post_processing_fn) -> Handler:
        logger = job_logger(getattr(event.json(), 'jobId', None))
        logger.debug("Create Qiskit Handler")

        request_data = event.json()
        provider_job_id = request_data.get("providerJobId")

        if provider_job_id is None:
            logger.debug('providerJobId is None. Create InvocationHandler')
            return InvocationHandler(request_data=request_data,
                                     circuit_preparation_fn=circuit_preparation_fn,
                                     post_processing_fn=post_processing_fn, )

        logger.debug('Create JobFetchingHandler')
        return JobFetchingHandler(request_data=request_data,
                                  post_processing_fn=post_processing_fn)



#  Quapp Platform Project
#  qiskit_circuit_export_task.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit import transpile
from quapp_common.async_tasks.export_circuit_task import CircuitExportTask
from quapp_common.enum.sdk import Sdk

from ..factory.qiskit_provider_factory import QiskitProviderFactory
from ..model.provider.oqc_cloud_provider import OqcCloudProvider


class QiskitCircuitExportTask(CircuitExportTask):

    def _transpile_circuit(self):
        self.logger.debug('Transpiling circuit')
        try:
            # Optional circuit quick stats
            try:
                num_qubits = getattr(self.circuit_data_holder.circuit,
                                     "num_qubits", None)
                depth = (self.circuit_data_holder.circuit.depth() if hasattr(
                        self.circuit_data_holder.circuit, "depth") else None)
                self.logger.debug(
                        f'Circuit stats: qubits={num_qubits}, depth={depth}')
            except Exception as exception:
                # Non-fatal: circuit might not support stats
                self.logger.warning(f'Failed to get circuit stats: {exception}')
                pass

            circuit = self.circuit_data_holder.circuit
            backend_information = self.backend_data_holder.backend_information

            self.logger.info(f'Creating provider: sdk={Sdk.QISKIT}, '
                             f'provider_tag={getattr(backend_information, "provider_tag", None)}')
            provider = QiskitProviderFactory.create_provider(sdk=Sdk.QISKIT,
                                                             provider_type=backend_information.provider_tag,
                                                             authentication=backend_information.authentication, )

            if isinstance(provider, OqcCloudProvider):
                self.logger.info(
                        f"Provider '{type(provider).__name__}' does not "
                        "require transpilation. Skipping.")
                return circuit

            device_name = getattr(backend_information, "device_name", None)
            self.logger.debug(f"Fetching backend '{device_name}'")
            backend = provider.get_backend(device_name)

            backend_name = getattr(backend, "name", None)
            try:
                backend_name = backend_name() if callable(
                        backend_name) else backend_name
            except Exception as exception:
                self.logger.warning(f'Failed to get backend name: {exception}')
                pass

            self.logger.info(
                    f"Starting transpilation for backend '{backend_name or device_name}'")
            transpiled = transpile(circuits=circuit, backend=backend)
            self.logger.info('Transpilation completed successfully')
            return transpiled

        except Exception as exception:
            self.logger.exception(f'Transpilation failed: {exception}')
            raise



#  Quapp Platform Project
#  qiskit_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from abc import ABC

from qiskit import QiskitError
from quapp_common.model.device.custom_device import CustomDevice


class QiskitDevice(CustomDevice, ABC):

    def _produce_histogram_data(self, job_result) -> dict | None:
        self.logger.debug('Producing histogram data')

        try:
            histogram_data = job_result.get_counts()
            self.logger.debug(f'Histogram data produced successfully: '
                              f'{histogram_data}')
            return histogram_data
        except QiskitError as qiskit_error:
            self.logger.debug(
                    f"Can't produce histogram data due to QiskitError: "
                    f"{str(qiskit_error)}. Returning None")
            return None
        except Exception as exception:
            self.logger.exception(
                    f'Unexpected error while producing histogram data: {exception}')
            return None

    def _get_provider_job_id(self, job) -> str:
        self.logger.debug('Get provider job id')
        try:
            job_id = job.job_id()
            self.logger.debug(f'Provider job id: {job_id}')
            return job_id
        except Exception as exception:
            self.logger.exception(f'Failed to get provider job id: {exception}')
            raise

    def _get_job_status(self, job) -> str:
        self.logger.debug('Get job status')

        try:
            status = str(job.status()).split(".")[-1]
            self.logger.debug(f'Job status resolved to: {status}')
            return status
        except Exception as exception:
            self.logger.exception(f'Failed to get job status: {exception}')
            raise

    def _calculate_execution_time(self, job_result):
        self.logger.debug('Calculate execution time')

        try:
            if "metadata" not in job_result:
                self.logger.debug("No 'metadata' key found in job_result")
                return None

            metadata = job_result["metadata"]
            if not metadata:
                self.logger.debug('Empty or None metadata in job_result')
                return None

            if "time_taken_execute" not in metadata:
                self.logger.debug(
                        "'time_taken_execute' not present in metadata")
                return None

            self.execution_time = metadata["time_taken_execute"]
            self.logger.debug(
                    f'Execution time calculation was: {self.execution_time} seconds')
            return None
        except Exception as exception:
            self.logger.exception(
                    f'Failed to calculate execution time: {exception}')
            return None

    def _get_job_result(self, job):
        self.logger.debug('Getting job result')
        try:
            result = job.result()
            self.logger.debug('Result fetched successfully')
            return result
        except Exception as exception:
            self.logger.exception(f'Failed to fetch job result: {exception}')
            raise

    def _get_shots(self, job_result) -> int | None:
        """
        Get the number of shots used in the job.

        :param job_result:
        :return: The number of shots used in the job if available, otherwise None.
        """
        try:
            if not job_result.results:
                self.logger.debug('No results found in job_result')
                return None

            first_result = job_result.results[0]
            if first_result is None:
                self.logger.debug('First result is None')
                return None

            shots = first_result.shots
            self.logger.debug(f'Number of shots: {shots}')
            return shots
        except Exception as exception:
            self.logger.exception(
                    f'Failed to get shots from job_result: {exception}')
            return None



#  Quapp Platform Project
#  job_fetching_handler.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.data.request.job_fetching_request import JobFetchingRequest
from quapp_common.handler.handler import Handler

from ..component.backend.qiskit_job_fetching import QiskitJobFetching


class JobFetchingHandler(Handler):
    def __init__(self, request_data: dict, post_processing_fn):
        super().__init__(request_data, post_processing_fn)

    def handle(self):
        self.logger.debug('Handle JobFetching: start')

        request = JobFetchingRequest(self.request_data)
        self.logger.debug(
                f'JobFetching request created successfully: {request.__dict__.keys()}')

        job_fetching = QiskitJobFetching(request)
        self.logger.info('Job fetching initiated')

        self.logger.debug('Job fetching started')
        fetching_result = job_fetching.fetch(
                post_processing_fn=self.post_processing_fn)
        self.logger.debug('Job fetching finished')

        return fetching_result



#  Quapp Platform Project
#  oqc_cloud_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from abc import ABC
from time import time

from qcaas_client.client import QPUTask, CompilerConfig
from qiskit.qasm2 import dumps
from quapp_common.config.logging_config import logger
from quapp_common.data.device.circuit_running_option import CircuitRunningOption
from quapp_common.enum.status.job_status import JobStatus
from quapp_common.model.device.custom_device import CustomDevice
from quapp_common.model.provider.provider import Provider


class OqcCloudDevice(CustomDevice, ABC):

    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)

        self.device_specification = device_specification

    def _create_job(self, circuit, options: CircuitRunningOption):
        self.logger.debug(f'Creating job with {options.shots} shots')

        start_time = time()

        qasm_str = dumps(circuit)
        circuit_submit_options = CompilerConfig(repeats=options.shots)

        self.logger.debug(f'QASM string: {qasm_str}')

        task = QPUTask(program=qasm_str, config=circuit_submit_options)
        self.logger.debug(f'Task created successfully: {task}')

        job = self.device.execute_tasks(task, qpu_id=self.device_specification)
        self.logger.debug(f'Job created successfully: {job}')

        self.execution_time = time() - start_time

        return job

    def _is_simulator(self) -> bool:
        self.logger.debug('Checking if simulator')
        self.logger.debug(f'Simulator: {True}')
        return True

    def _produce_histogram_data(self, job_result) -> dict | None:
        self.logger.debug('Producing histogram data')
        histogram = next(iter(job_result.result.values()))
        self.logger.debug(f'Histogram data produced successfully: {histogram}')
        return histogram

    def _get_provider_job_id(self, job) -> str:
        self.logger.debug('Get provider job id')
        provider_job_id = job[0].id
        self.logger.debug(f'Provider job id: {provider_job_id}')
        return provider_job_id

    def _get_job_status(self, job) -> str:
        self.logger.debug('Get job status')
        oqc_status = self.device.get_task_status(task_id=job[0].id,
                                                 qpu_id=self.device_specification)
        self.logger.debug(f'Job status resolved to: {oqc_status}')

        if "FAILED".__eq__(oqc_status):
            self.logger.info("Job failed")
            return JobStatus.ERROR.value
        elif "COMPLETED".__eq__(oqc_status):
            self.logger.info("Job completed")
            return JobStatus.DONE.value

        self.logger.info("Job is still running")
        return oqc_status

    def _calculate_execution_time(self, job_result):
        self.logger.debug('Calculate execution time')
        self.logger.debug(
                f'Execution time calculation was: {self.execution_time} seconds')

    def _get_job_result(self, job):
        self.logger.debug('Getting job result')
        job_result = job[0]
        self.logger.info(f'Job result fetched successfully: {job_result}')
        return job_result

    def _get_shots(self, job_result) -> int | None:
        self.logger.debug('Getting shots')
        result_classical = job_result.result.get('classical')
        if result_classical is not None:
            logger.debug(f'Classical results found: {result_classical}')
            total_shots = sum(value for value in result_classical.values())
            self.logger.debug(f'Total shots: {total_shots}')
            return total_shots

        logger.debug('No classical results found.')
        return None



"""
    QApp Platform Project ibm_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  ibm_cloud_provider.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit_ibm_runtime.accounts import ChannelType
from quapp_common.config.logging_config import job_logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

logger = job_logger('IbmCloudProvider')


class IbmCloudProvider(Provider):

    def __init__(self, api_key, crn):
        super().__init__(ProviderTag.IBM_CLOUD)
        self.api_key = api_key
        self.crn = crn
        self.channel: ChannelType = "ibm_cloud"

    def get_backend(self, device_specification: str):
        logger.debug('Get backend from IBM Cloud')

        provider = self.collect_provider()

        backend = provider.backend(name=device_specification)
        logger.debug(f'Backend: {backend}')
        return backend

    def collect_provider(self):
        logger.debug('Collecting provider for IBM Cloud')

        return QiskitRuntimeService(channel=self.channel, token=self.api_key,
                                    instance=self.crn)



from dataclasses import dataclass

@dataclass
class RikenEndpoint:
  submit_job: str
  job_status: str
  job_cancel: str
  job_delete: str
  job_list: str

@dataclass
class RikenConfig:
  endpoints: RikenEndpoint
  client_id: str
  client_secret: str
  token_endpoint: str
  qc_token: str



#  Quapp Platform Project
#  qiskit_provider_factory.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk
from quapp_common.factory.provider_factory import ProviderFactory

from ..model.provider.riken_provider import RikenProvider
from ..model.provider.data import RikenConfig, RikenEndpoint
from ..model.provider.ibm_cloud_provider import IbmCloudProvider
from ..model.provider.ibm_quantum_provider import IbmQuantumProvider
from ..model.provider.oqc_cloud_provider import OqcCloudProvider
from ..model.provider.qapp_qiskit_provider import QappQiskitProvider


class QiskitProviderFactory(ProviderFactory):

    @staticmethod
    def create_provider(provider_type: ProviderTag, sdk: Sdk,
            authentication: dict):
        logger.debug("Create Qiskit Provider")

        if ProviderTag.QUAO_QUANTUM_SIMULATOR.__eq__(
                provider_type) and Sdk.QISKIT.__eq__(sdk):
            logger.debug("Create Qiskit Qapp Qiskit Provider")
            return QappQiskitProvider()

        if ProviderTag.IBM_QUANTUM.__eq__(provider_type):
            logger.debug("Create IBM Quantum Provider")
            return IbmQuantumProvider(authentication.get("token"))

        if ProviderTag.IBM_CLOUD.__eq__(provider_type):
            logger.debug("Create IBM Cloud Provider")
            return IbmCloudProvider(authentication.get("token"),
                    authentication.get("crn"))

        if ProviderTag.OQC_CLOUD.__eq__(provider_type):
            logger.debug("Create OQC Cloud Provider")
            return OqcCloudProvider(authentication.get("url"),
                    authentication.get("accessToken"))

        if ProviderTag.RIKEN.__eq__(provider_type):
            return RikenProvider(config=RikenConfig(
                    client_id=authentication.get("clientId"),
                    client_secret=authentication.get("clientSecret"),
                    token_endpoint=authentication.get("tokenEndpoint"),
                    qc_token=authentication.get("qcToken"),
                    endpoints=RikenEndpoint(
                        submit_job=authentication.get("endpoints").get("submitJob"),
                        job_list=authentication.get("endpoints").get("jobList"),
                        job_status=authentication.get("endpoints").get("jobStatus"),
                        job_cancel=authentication.get("endpoints").get("jobCancel"),
                        job_delete=authentication.get("endpoints").get("jobDelete"),
                    ))
                , qc_type=authentication.get("metadata").get("qc_type")
            )
        raise Exception("Unsupported provider!")



#  Quapp Platform Project
#  invocation_handler.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.data.request.invocation_request import InvocationRequest
from quapp_common.handler.handler import Handler

from ..component.backend.qiskit_invocation import QiskitInvocation


class InvocationHandler(Handler):
    def __init__(self, request_data: dict, circuit_preparation_fn,
                 post_processing_fn):
        super().__init__(request_data, post_processing_fn)
        self.circuit_preparation_fn = circuit_preparation_fn

    def handle(self):
        self.logger.debug('Handle Invocation: start')
        try:

            invocation_request = InvocationRequest(self.request_data)
            self.logger.debug(f'Invocation request created successfully: '
                              f'{invocation_request.__dict__.keys()}')

            backend = QiskitInvocation(invocation_request)
            self.logger.info('Submitting job via QiskitInvocation')

            backend.submit_job(
                    circuit_preparation_fn=self.circuit_preparation_fn,
                    post_processing_fn=self.post_processing_fn)

            self.logger.debug('Job submission triggered successfully')
        except Exception as exception:
            self.logger.exception(f'Invocation handling failed: {exception}')
            raise



#  Quapp Platform Project
#  qiskit_invocation.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit import QuantumCircuit
from quapp_common.component.backend.invocation import Invocation
from quapp_common.config.thread_config import circuit_exporting_pool
from quapp_common.data.async_task.circuit_export.backend_holder import \
    BackendDataHolder
from quapp_common.data.async_task.circuit_export.circuit_holder import \
    CircuitDataHolder
from quapp_common.data.request.invocation_request import InvocationRequest
from quapp_common.model.provider.provider import Provider

from ...async_tasks.qiskit_circuit_export_task import QiskitCircuitExportTask
from ...factory.qiskit_device_factory import QiskitDeviceFactory
from ...factory.qiskit_provider_factory import QiskitProviderFactory


class QiskitInvocation(Invocation):

    def __init__(self, request_data: InvocationRequest):
        super().__init__(request_data)

    def _export_circuit(self, circuit):
        self.logger.debug('Exporting circuit to backend')
        try:
            # Circuit quick stats (best-effort)
            try:
                num_qubits = getattr(circuit, "num_qubits", None)
                depth = circuit.depth() if hasattr(circuit, "depth") else None
                self.logger.debug(
                        f'Circuit stats: qubits={num_qubits}, depth={depth}')

            except Exception as exception:
                # Non-fatal diagnostics
                self.logger.warning(f'Failed to get circuit stats: {exception}')
                pass

            # Backend info (avoid logging sensitive details)
            provider_tag = getattr(self.backend_information, "provider_tag",
                                   None)
            device_name = getattr(self.backend_information, "device_name", None)
            self.logger.debug(
                    f'Backend info: provider_tag={provider_tag}, device_name={device_name}')

            circuit_export_task = QiskitCircuitExportTask(
                    circuit_data_holder=CircuitDataHolder(circuit,
                                                          self.circuit_export_url),
                    backend_data_holder=BackendDataHolder(
                            self.backend_information,
                            self.authentication.user_token),
                    project_header=self.project_header,
                    workspace_header=self.workspace_header)

            self.logger.info('Submitting circuit export task to thread pool')
            circuit_exporting_pool.submit(circuit_export_task.do)
            self.logger.debug('Circuit export task submitted successfully')

        except Exception as exception:
            self.logger.exception(f'Circuit export failed: {exception}')
            raise

    def _create_provider(self):
        try:
            self.logger.debug(
                    f'Creating provider: provider_tag={getattr(self.backend_information, "provider_tag", None)}, sdk={getattr(self, "sdk", None)}')
            provider = QiskitProviderFactory.create_provider(
                    provider_type=self.backend_information.provider_tag,
                    sdk=self.sdk,
                    authentication=self.backend_information.authentication, )
            self.logger.info(
                    f'Provider created successfully: {type(provider).__name__}')
            return provider
        except Exception as exception:
            self.logger.exception(f'Provider creation failed: {exception}')
            raise

    def _create_device(self, provider: Provider):
        self.logger.debug('Creating device')
        try:
            device_spec = getattr(self.backend_information, "device_name", None)
            self.logger.debug(
                    f'Creating device from provider={type(provider).__name__}, '
                    f'device_specification={device_spec}, sdk={getattr(self, "sdk", None)}', )
            device = QiskitDeviceFactory.create_device(provider=provider,
                                                       device_specification=device_spec,
                                                       authentication=self.backend_information.authentication,
                                                       sdk=self.sdk)
            self.logger.info(
                    f'Device created successfully: {type(device).__name__}')
            return device
        except Exception as exception:
            self.logger.exception(f"Device creation failed: {exception}")
            raise

    def _get_qubit_amount(self, circuit):
        self.logger.debug(
                f'Getting qubit amount for circuit type: {type(circuit).__name__}')
        if isinstance(circuit, QuantumCircuit):
            qubits = int(circuit.num_qubits)
            self.logger.debug(f'Qubit amount determined: {qubits}')
            return qubits

        self.logger.exception(
                f'Invalid circuit type for qubit amount: {type(circuit).__name__}')
        raise Exception("Invalid circuit type!")



"""
    QApp Platform Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  ibm_cloud_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

import time
from qiskit import qasm2
import base64

from quapp_common.data.device.circuit_running_option import CircuitRunningOption


from .qiskit_device import QiskitDevice


class RikenDevice(QiskitDevice):
    def _is_simulator(self) -> bool:
        return True

    def _create_job(self, circuit, options: CircuitRunningOption):

        qasm = qasm2.dumps(circuit)
        riken_response = self.device.submit_job(qasm,options.shots)

        job_id = riken_response.get("jobId")

        while True:
            status_response = self.device.get_job_status(job_id)
            current_status = status_response.get("status")

            if current_status != 4 and current_status != 5 and current_status != 6:
                self.logger.info(f"Job status: {current_status}. Waiting for completion...")
                time.sleep(5)
            else:
                self.logger.info(
                    f"Job ended, status: {current_status}.")
                return status_response



    def _get_job_status(self, job) -> str:
        return "DONE"

    def _calculate_execution_time(self, job_result):
        # not implemented
        pass

    def _produce_histogram_data(self, job_result) -> dict | None:
        return None

    def _get_provider_job_id(self, job) -> str:
        self.logger.debug('Get provider job id')
        return job.get("qcJobId")


    def _get_job_result(self, job):
        return job.get("result")

    def _get_shots(self, job_result) -> int | None:
        return None



"""
    QApp Platform Project ibm_quantum_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  ibm_quantum_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit import transpile
from qiskit_ibm_runtime import SamplerV2 as Sampler
from quapp_common.data.device.circuit_running_option import CircuitRunningOption

from .ibm_cloud_device import IbmCloudDevice


class IbmQuantumDevice(IbmCloudDevice):

    def _create_job(self, circuit, options: CircuitRunningOption):
        self.logger.debug(f"Creating job with {options.shots} shots")

        transpiled_circuit = transpile(circuits=circuit, backend=self.device)
        self.logger.debug('Transpiled circuit created successfully')

        sampler = Sampler(self.device)
        self.logger.debug('Sampler created successfully')

        job = sampler.run([transpiled_circuit], shots=options.shots)
        self.logger.info('Job submitted successfully')
        return job



"""
    QApp Platform Project ibm_cloud_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  ibm_cloud_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit_ibm_runtime import SamplerV2 as Sampler
from quapp_common.data.device.circuit_running_option import CircuitRunningOption
from qiskit import transpile
import time


from .qiskit_device import QiskitDevice


class IbmCloudDevice(QiskitDevice):
    def _is_simulator(self) -> bool:
        self.logger.debug('Checking if simulator')
        simulator = self.device.configuration().simulator
        self.logger.debug(f'Simulator: {simulator}')
        return True

    def _create_job(self, circuit, options: CircuitRunningOption):
        self.logger.debug(f"Creating job with {options.shots} shots")

        transpiled_circuit = transpile(circuits=circuit, backend=self.device)
        self.logger.debug('Transpiled circuit created successfully')

        sampler = Sampler(self.device)
        self.logger.debug('Sampler created successfully')

        job = sampler.run([transpiled_circuit], shots=options.shots)
        self.logger.info('Job submitted successfully')

        while True:
            current_status = job.status()
            if current_status != "ERROR" and current_status != "DONE" and current_status != "CANCELLED":
                self.logger.info(f"Job status: {current_status}. Waiting for completion...")
                time.sleep(5)
            else:
                self.logger.info(
                    f"Job ended, status: {current_status}.")
                break

        self.execution_time = job.usage()  # Seconds

        return job
    def _get_job_status(self, job) -> str:
        return "DONE"

    def _calculate_execution_time(self, job_result):
        # calculate execution time is done in _create_job for IBM Cloud
        pass


"""
    QApp Platform Project oqc_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  oqc_cloud_provider.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qcaas_client.client import OQCClient
from quapp_common.config.logging_config import job_logger

from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

logger = job_logger('OqcCloudProvider')


class OqcCloudProvider(Provider):
    def __init__(self, oqc_cloud_url: str, access_token: str):
        super().__init__(ProviderTag.OQC_CLOUD)
        self.oqc_cloud_url = oqc_cloud_url
        self.access_token = access_token

    def get_backend(self, device_specification):
        logger.debug('Get backend from OQC Cloud')

        return self.collect_provider()

    def collect_provider(self):
        logger.debug('Collecting provider for OQC Cloud')
        return OQCClient(url=self.oqc_cloud_url,
                         authentication_token=self.access_token)



"""
    QApp Platform Project qapp_qiskit_device.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  qapp_qiskit_device.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit import transpile
from quapp_common.data.device.circuit_running_option import CircuitRunningOption
from quapp_common.model.provider.provider import Provider

from .qiskit_device import QiskitDevice


class QappQiskitDevice(QiskitDevice):
    def __init__(self, provider: Provider, device_specification: str):
        super().__init__(provider, device_specification)

    def _create_job(self, circuit, options: CircuitRunningOption):
        self.logger.debug(f'Creating job with {options.shots} shots')
        self.device.set_options(device=options.processing_unit.value,
                                shots=options.shots, executor=options.executor,
                                max_job_size=options.max_job_size)
        self.logger.debug(
                f'Set options: device={options.processing_unit.value}, '
                f'shots={options.shots}, executor={options.executor}, '
                f'max_job_size={options.max_job_size}')

        transpiled_circuit = transpile(circuits=circuit, backend=self.device)
        self.logger.debug('Transpiled circuit created successfully')

        return self.device.run(transpiled_circuit)

    def _is_simulator(self) -> bool:
        self.logger.debug('Checking if simulator')
        self.logger.debug(f'Simulator: {True}')
        return True



"""
    QApp Platform Project qapp_qiskit_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  qapp_qiskit_provider.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit_aer import Aer
from quapp_common.config.logging_config import job_logger

from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

logger = job_logger('QappQiskitProvider')


class QappQiskitProvider(Provider):

    def __init__(self, ):
        super().__init__(ProviderTag.QUAO_QUANTUM_SIMULATOR)

    def get_backend(self, device_specification):
        logger.debug('Get backend from Qapp Qiskit')

        provider = self.collect_provider()

        device_names = {self.__map_aer_backend_name(backend_name) for
                        backend_name in provider.backends()}

        if device_names.__contains__(device_specification):
            backend = provider.get_backend(device_specification)
            logger.debug(f'Backend: {backend}')
            return backend

        logger.exception(f'Unsupported device: {device_specification}')
        raise Exception('Unsupported device')

    def collect_provider(self):
        logger.debug('Collecting provider for Qapp Qiskit')
        return Aer

    @staticmethod
    def __map_aer_backend_name(backend):
        return backend.configuration().backend_name



#  Quapp Platform Project
#  qiskit_device_factory.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.config.logging_config import logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk
from quapp_common.factory.device_factory import DeviceFactory
from quapp_common.model.provider.provider import Provider

from ..model.device.ibm_cloud_device import IbmCloudDevice
from ..model.device.ibm_quantum_device import IbmQuantumDevice

from ..model.device.riken_device import RikenDevice
from ..model.device.oqc_cloud_device import OqcCloudDevice
from ..model.device.qapp_qiskit_device import QappQiskitDevice
from quapp_common.config.logging_config import logger


class QiskitDeviceFactory(DeviceFactory):

    @staticmethod
    def create_device(provider: Provider, device_specification: str,
            authentication: dict, sdk: Sdk):
        logger.info("[QiskitDeviceFactory] create_device()")

        provider_type = ProviderTag.resolve(provider.get_provider_type().value)



        if ProviderTag.QUAO_QUANTUM_SIMULATOR.__eq__(
                provider_type) and Sdk.QISKIT.__eq__(sdk):
            logger.info("Created QUAPP_QUANTUM_SIMULATOR device")
            return QappQiskitDevice(provider, device_specification)

        if ProviderTag.IBM_QUANTUM.__eq__(provider_type):
            logger.info("Created IBM_QUANTUM device")
            return IbmQuantumDevice(provider, device_specification)

        if ProviderTag.IBM_CLOUD.__eq__(provider_type):
            logger.info("Created IBM_CLOUD device")
            return IbmCloudDevice(provider, device_specification)

        if ProviderTag.RIKEN.__eq__(provider_type):
            logger.info("Created RIKEN device")
            return RikenDevice(provider, device_specification)

        if ProviderTag.OQC_CLOUD.__eq__(provider_type):
            logger.info("Created OQC_CLOUD device")
            return OqcCloudDevice(provider, device_specification)

        raise Exception("Unsupported device!")



"""
    QApp Platform Project ibm_cloud_provider.py Copyright © CITYNOW Co. Ltd. All rights reserved.
"""
#  Quapp Platform Project
#  ibm_cloud_provider.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit_ibm_runtime.accounts import ChannelType

from quapp_common.config.logging_config import job_logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider
import requests
from requests.auth import HTTPBasicAuth
from .data import RikenConfig

logger = job_logger('RikenProvider')


class RikenProvider(Provider):

    def collect_provider(self):
        #Not applicable for RikenProvider
        pass

    def __init__(self, config: RikenConfig, qc_type: int):
        super().__init__(ProviderTag.RIKEN)
        self.config = config
        self.qc_type = qc_type

    def get_backend(self, device_specification: str):
        logger.debug('Get backend from Riken')



        return RikenBackend(
            config= self.config,
            device_specification=device_specification,
            qc_type=self.qc_type
        )

class RikenBackend:
    def __init__(self,
        config: RikenConfig,
        device_specification: str = None,
        qc_type: int = 99999):
        self.config = config
        self.device_specification = device_specification
        self.qc_type = qc_type

    def get_token(self):
        data = {
            'grant_type': 'client_credentials',
            'scope': 'apiauth/read'
        }

        # 2. Gọi API Token Endpoint
        try:
            auth = HTTPBasicAuth(self.config.client_id, self.config.client_secret)

            response = requests.post(
                self.config.token_endpoint,
                data=data,
                auth=auth,
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                timeout=10
            )

            response.raise_for_status()

            # 3. Parse kết quả
            token_data = response.json()

            # Lấy access_token
            access_token = token_data.get('access_token')

            if not access_token:
                raise ValueError("No access_token")

            return access_token

        except requests.exceptions.RequestException as e:
            logger.error(e)
            raise e

    def submit_job(self, qasm_base64: str,
        shots: int,
        priority: int = 1):

        payload = {
            "token": self.config.qc_token,
            "priority": priority,
            "qasm": qasm_base64,
            "shots": shots,
            "qc_type": self.qc_type
        }

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.get_token()}"
        }

        try:
            response = requests.post(self.config.endpoints.submit_job, json=payload, headers=headers,
                                     timeout=30)
            response.raise_for_status()

            return response.json()

        except requests.exceptions.RequestException as e:
            raise RuntimeError(e)

    def get_job_status(self, job_id: str):

        payload = {
            "token": self.config.qc_token,
            "job_id": job_id
        }

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.get_token()}"
        }

        try:
            response = requests.post(self.config.endpoints.job_status, json=payload, headers=headers,
                                     timeout=30)
            response.raise_for_status()

            return response.json()

        except requests.exceptions.RequestException as e:
            raise RuntimeError(e)


#  Quapp Platform Project
#  ibm_quantum_provider.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from qiskit_ibm_runtime import QiskitRuntimeService
from quapp_common.config.logging_config import job_logger
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.model.provider.provider import Provider

logger = job_logger('IbmQuantumProvider')


class IbmQuantumProvider(Provider):

    def __init__(self, api_token):
        super().__init__(ProviderTag.IBM_QUANTUM)
        self.api_token = api_token

    def get_backend(self, device_specification: str):
        logger.debug('Get backend from IBM Quantum')

        provider = self.collect_provider()

        backend = provider.backend(device_specification)
        logger.debug(f'Backend: {backend}')
        return backend

    def collect_provider(self):
        logger.debug('Collecting provider for IBM Quantum')

        return QiskitRuntimeService(channel='ibm_quantum', token=self.api_token)



#  Quapp Platform Project
#  qiskit_job_fetching.py
#  Copyright © CITYNOW Co. Ltd. All rights reserved.

from quapp_common.component.backend.job_fetching import JobFetching
from quapp_common.data.request.job_fetching_request import JobFetchingRequest
from quapp_common.enum.provider_tag import ProviderTag
from quapp_common.enum.sdk import Sdk

from ...factory.qiskit_provider_factory import QiskitProviderFactory


class QiskitJobFetching(JobFetching):

    def __init__(self, request_data: JobFetchingRequest):
        super().__init__(request_data)

    def _collect_provider(self, ):
        try:
            self.logger.debug(
                    f'Collecting provider with provider_type={ProviderTag.IBM_QUANTUM}, '
                    f'sdk={Sdk.QISKIT}')
            provider = QiskitProviderFactory.create_provider(
                    provider_type=ProviderTag.IBM_QUANTUM, sdk=Sdk.QISKIT,
                    authentication=self.provider_authentication, ).collect_provider()
            self.logger.info('Provider collected successfully')
            return provider
        except Exception as exception:
            self.logger.exception(f'Failed to collect provider: {exception}')
            raise

    def _retrieve_job(self, provider):
        try:
            self.logger.debug(
                    f'Retrieving job with job_id={self.provider_job_id}')
            job = provider.job(job_id=self.provider_job_id)
            self.logger.info('Job retrieved successfully')
            return job
        except Exception as exception:
            self.logger.exception(f'Failed to retrieve job '
                                  f'{self.provider_job_id}: {exception}')
            raise

    def _get_job_status(self, job):
        self.logger.debug('Getting job status')
        try:
            status = job.status()
            self.logger.debug(f'Job status: {status}')
            return status
        except Exception as exception:
            self.logger.exception(f'Failed to get job status for job_id='
                                  f'{self.provider_job_id}: '
                                  f'{exception}')
            raise

    def _get_job_result(self, job):
        self.logger.info("_get_job_result()")
        try:
            self.logger.debug("Returning raw job object for job_id=%s",
                              self.provider_job_id)
            return job
        except Exception as exception:
            self.logger.exception("Failed to get job result for job_id=%s: %s",
                                  self.provider_job_id, e)
            raise
